from dataclasses import dataclass

TimeoutSecond = 5 # 5s for timeout

Bili = [
    "tv.danmaku.bili", # Bilibili CN
    "com.bilibili.app.in", # Bilibili Google Play
]

PrimaryRoot = "/storage/emulated/0"
AndroidData = "/storage/emulated/0/Android/data"
EntryJson = "entry.json"
AudioFile = "audio.m4s"
Music = "/storage/emulated/0/Music"

@dataclass
class Entry:
    title: str
    type_tag: str
    total_time_milli: int
    owner_id: int # CN does not have owner_name
    cid: int
    # part: str # CN do not support it



from abc import ABC, abstractmethod
from contextlib import contextmanager


class IDevice(ABC):
    @classmethod
    def create(cls, host: str, port: int) -> "IDevice":
        return cls(host, port)

    @abstractmethod
    def run(self, cmd: str) -> str:
        pass

    @abstractmethod
    def close(self) -> None:
        pass


DEVICES: dict[str, IDevice] = {}
DEFAULT_DEVICE: list[str] = []


def register_device(name: str, default: bool = False):
    def wrapper(cls):
        DEVICES[name] = cls
        if default:
            if len(DEFAULT_DEVICE) == 0:
                DEFAULT_DEVICE.append(name)
            else:
                print(
                    f"[Warning] default device changed: {DEFAULT_DEVICE[0]} -> {name}"
                )
                DEFAULT_DEVICE[0] = name
        elif len(DEFAULT_DEVICE) == 0:
            DEFAULT_DEVICE.append(name)
        return cls

    return wrapper


@contextmanager
def open(host: str, port: int):
    try:
        device = DEVICES[DEFAULT_DEVICE[0]].create(host, port)
        yield device
    finally:
        device.close()



try:
    import adbutils
except ImportError:
    print("pip install adbutils")
    raise

from adbili.devices import IDevice, register_device

# @DeprecationWarning("Use AdbCliDevice instead")
@register_device("atx")
class AtxDevice(IDevice):
    """
    AdbDevice from openatx/adbutils
    """

    def __init__(self, host: str, port: int):
        self.addr = f"{host}:{port}"
        self.adb = adbutils.AdbClient(host=host, port=port)
        self.device = self.adb.device()

    def run(self, cmd: str) -> str:
        return self.device.shell(cmd)

    def close(self) -> None:
        self.adb.disconnect(self.addr)



import subprocess
import os

from adbili.devices import IDevice, register_device


@register_device("adb", default=True)
class AdbCliDevice(IDevice):
    """
    delegate to adb cli
    """

    def __init__(self, host: str, port: int):
        pass

    def run(self, cmd: str) -> str:
        result = subprocess.run(f"adb shell '{cmd}'", shell=True, capture_output=True)
        if result.stderr:
            print(result.stderr.decode("utf-8"))
        return result.stdout.decode("utf-8")
        # return os.system(f"adb shell '{cmd}'")

    def close(self) -> None:
        pass



from .constant import AndroidData, EntryJson, AudioFile


def app_exist(appid: str):
    return f"[ -e {AndroidData}/{appid} ] || echo NO"


def list_download(appid: str):
    return f"ls {AndroidData}/{appid}/download"


def list_cid(appid: str, avid: str):
    return f"ls {AndroidData}/{appid}/download/{avid}"


def get_entry_json(appid: str, avid: str, cid: str):
    return f"{AndroidData}/{appid}/{avid}/{cid}/{EntryJson}"


def collect_entries(appid: str):
    return f"ls {AndroidData}/{appid}/download/*/*/{EntryJson}"

def get_audio_m4s(entry: str, tag: str):
    entry_dir = entry[: -len(EntryJson)]
    audio_file = f"{entry_dir}{tag}/{AudioFile}"
    return audio_file



import readline

from adbili.constant import Entry

def time_range_filter(entry: Entry, min_sec:int, max_sec:int,) -> bool:
    return min_sec < entry.total_time_milli/1000 < max_sec

def create_entry_from_json(js: dict) -> Entry:
    return Entry(
        title=js["title"],
        type_tag=js["type_tag"],
        total_time_milli=js["total_time_milli"],
        owner_id=js["owner_id"],
        cid=js["page_data"]["cid"],
        # part=js["page_data"]["part"], # for Google Play
    )


def get_line(prompt: str) -> str:
    return input(prompt)



import json
from typing import Callable
from functools import partial

from adbili.constant import Bili, Music, Entry
from adbili.command import app_exist, collect_entries, get_audio_m4s, list_download, list_cid
from adbili.utils import (
    create_entry_from_json,
    get_line,
    time_range_filter,
)
from adbili import devices


def process_app(
    device: devices.IDevice,
    appid: str,
    target: str,
    prompt: bool,
    recursive: bool,
    filter_partial: Callable[[Entry], bool],
):
    cmd_list_download = list_download(appid)
    items = device.run(cmd_list_download).splitlines()
    # breakpoint()
    for avid in items:
        cmd_ls_item = list_cid(appid, avid)
        item_entry = device.run(cmd_ls_item).splitlines()
        if len(item_entry) > 1 and not recursive:
            continue
        for ep in item_entry:
            entry_json_file = f"{cmd_ls_item.split()[-1]}/{ep}/entry.json"
            entry_json = device.run(f"cat {entry_json_file}")
            entry = create_entry_from_json(json.loads(entry_json))
            audio_file = get_audio_m4s(entry_json_file, entry.type_tag)
            if not filter_partial(entry):
                continue  # apply filter
            dest = entry.title.strip()
            if prompt:
                opt = get_line(f"audio name: [{entry.title}] (Press , to skip) ")
                if opt == ",":
                    continue
                elif not opt.isspace():
                    dest = opt
            cp_cmd = f"cp {audio_file} {target}/{dest}.mp3"
            device.run(cp_cmd)
            print("[Done]", cp_cmd)

def app(
    host: str = "",
    port: int = 0,
    target: str = Music,
    prompt: bool = True,
    app: str | None = None,  # app filter
    recursive: bool = False, # allow video collection
):
    assert app in Bili, f"Unknown appId: {app}. It is not in {Bili}"
    with devices.open(host, port) as device:
        result = device.run(app_exist(app))
        if result:
            print(f"[{app}] not exists")
            return
        process_app(
            device,
            app,
            target,
            prompt,
            recursive,
            partial(time_range_filter, min_sec=1 * 60, max_sec=5 * 60),
        )
